# pylint: disable=too-many-arguments
import os
import sys
from pathlib import Path

import click
import orjson
import requests

from seagoat import __version__
from seagoat.utils.cli_display import display_results
from seagoat.utils.config import get_config_values
from seagoat.utils.server import get_server_info
from seagoat.utils.server import ServerDoesNotExist


# pylint: disable-next=too-few-public-methods
class ExitCode:
    SERVER_NOT_RUNNING = 3
    SERVER_ERROR = 4


def warn_if_update_available():
    response = requests.get("https://pypi.org/pypi/seagoat/json")
    latest_version = orjson.loads(response.text)["info"]["version"]
    if latest_version != __version__:
        click.echo(
            f"Warning: An updated version {latest_version} of SeaGOAT is available. You have {__version__}.",
            err=True,
        )


def display_accuracy_warning(server_address):
    response = requests.get(
        f"{server_address}/status",
    )
    response_data = orjson.loads(response.text)
    accuracy = response_data["stats"]["accuracy"]["percentage"]

    if accuracy < 100:
        click.echo(
            click.style(
                "Warning: SeaGOAT is still analyzing your repository. "
                + f"The results displayed have an estimated accuracy of {accuracy}%",
                fg="red",
            ),
            err=True,
        )


def query_server(query, server_address, max_results, context_above, context_below):
    response = requests.get(
        f"{server_address}/query/{query}",
        params={
            "limitClue": max_results,
            "contextAbove": context_above,
            "contextBelow": context_below,
        },
    )

    response_data = orjson.loads(response.text)
    if "error" in response_data:
        click.echo(response_data["error"]["message"], err=True)
        sys.exit(ExitCode.SERVER_ERROR)

    response.raise_for_status()

    return response_data["results"]


def rewrite_full_paths_to_use_local_path(repo_path, results):
    return [
        {
            **result,
            "fullPath": str((Path(repo_path) / result["path"]).expanduser().resolve()),
        }
        for result in results
    ]


def remove_results_from_unavailable_files(results):
    return [result for result in results if Path(result["fullPath"]).exists()]


@click.command()
@click.argument("query")
@click.argument("repo_path", required=False, default=os.getcwd())
@click.option(
    "--no-color",
    is_flag=True,
    help="Disable formatting. Automatically enabled when part of a bash pipeline.",
)
@click.option(
    "-l",
    "--max-results",
    type=int,
    default=None,
    help="Limit the number of result lines",
)
@click.option(
    "-B",
    "--context-above",
    type=int,
    default=None,
    help="Include this many lines of context before each result",
)
@click.option(
    "-A",
    "--context-below",
    type=int,
    default=None,
    help="Include this many lines of context after each result",
)
@click.option(
    "-C",
    "--context",
    type=int,
    default=None,
    help="Include this many lines of context after and before each result",
)
@click.version_option(version=__version__, prog_name="seagoat")
def seagoat(
    query, repo_path, no_color, max_results, context_above, context_below, context
):
    """
    Query your codebase for your QUERY in the Git repository REPO_PATH.
    Your query can contain keywords, regular expression patterns,
    or a description of what you are looking for.

    When REPO_PATH is not specified, the current working directory is
    assumed to be the repository path.

    In order to use seagoat in your repository, you need to run a server
    that will analyze your codebase. Check seagoat-server --help for more details.
    """
    config = get_config_values(Path(repo_path))

    try:
        if config["client"]["host"] is None:
            server_info = get_server_info(repo_path)
            server_address = server_info["address"]
        else:
            server_address = config["client"]["host"]

        if context is not None:
            context_above = context
            context_below = context

        results = query_server(
            query,
            server_address,
            max_results,
            context_above or 0,
            context_below or 0,
        )

        results = rewrite_full_paths_to_use_local_path(repo_path, results)
        results = remove_results_from_unavailable_files(results)

        color_enabled = os.isatty(0) and not no_color

        display_results(results, max_results, color_enabled)

        display_accuracy_warning(server_address)
        warn_if_update_available()
    except (
        requests.exceptions.ConnectionError,
        requests.exceptions.RequestException,
        ServerDoesNotExist,
    ):
        click.echo(
            f"The SeaGOAT server is not running. "
            f"Please start the server using the following command:\n\n"
            f"seagoat-server start {repo_path}\n"
        )
        sys.exit(ExitCode.SERVER_NOT_RUNNING)


if __name__ == "__main__":
    # pylint: disable-next=no-value-for-parameter
    seagoat()



# pylint: disable=too-few-public-methods
import functools
import re
from collections import Counter
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict
from typing import List
from typing import Set

from seagoat.utils.file_reader import read_file_with_correct_encoding
from seagoat.utils.file_types import get_file_penalty_factor


SPLITTER_PATTERN = re.compile(r"\s+")


@functools.cache
def get_number_of_exact_matches(line: str, query: str):
    terms = re.split(SPLITTER_PATTERN, query)
    pattern = ".*".join(map(re.escape, terms))

    if re.search(pattern, line, re.IGNORECASE):
        return 1
    return 0


class ResultLineType(Enum):
    RESULT = "result"
    CONTEXT = "context"

    def __str__(self):
        return self.value


@dataclass(frozen=True)
class ResultLine:
    line: int
    vector_distance: float
    line_text: str
    types: Set[ResultLineType]

    def get_score(self, query: str) -> float:
        return self.vector_distance / (
            1 + get_number_of_exact_matches(self.line_text, query)
        )

    def add_type(self, type_: ResultLineType) -> None:
        self.types.add(type_)

    def to_json(self, query: str):
        return {
            "score": round(self.get_score(query), 4),
            "line": self.line,
            "lineText": self.line_text,
            "resultTypes": list(sorted(set(str(t) for t in self.types))),
        }


@dataclass(frozen=True)
class ResultBlock:
    lines: List[ResultLine]

    def _get_line_count_per_type(self) -> Dict[str, int]:
        counts = Counter()

        for line in self.lines:
            for type_ in line.types:
                counts[str(type_)] += 1

        return dict(counts)

    def to_json(self, query: str):
        return {
            "lines": [line.to_json(query) for line in self.lines],
            "lineTypeCount": self._get_line_count_per_type(),
        }


class Result:
    def __init__(self, path: str, full_path: Path) -> None:
        self.path: str = path
        self.full_path: Path = full_path
        self.lines: Dict[int, ResultLine] = {}
        self.line_texts = read_file_with_correct_encoding(self.full_path).splitlines()

    def __repr__(self) -> str:
        return f"Result(path={self.path})"

    def extend(self, other) -> None:
        self.lines.update(other.lines)

    def add_line(self, line: int, vector_distance: float) -> None:
        types = set()
        if line in self.lines and self.lines[line].vector_distance < vector_distance:
            vector_distance = self.lines[line].vector_distance
            types = self.lines[line].types

        types.add(
            ResultLineType.RESULT,
        )

        self.lines[line] = ResultLine(
            line,
            vector_distance,
            self.line_texts[line - 1],
            types,
        )

    def get_best_score(self, query: str) -> float:
        best_score = min(
            (x for x in self.lines.values() if ResultLineType.RESULT in x.types),
            key=lambda item: item.get_score(query),
        ).get_score(query)

        best_score *= get_file_penalty_factor(self.full_path)

        return best_score

    def get_lines(self, query: str):
        best_score = self.get_best_score(query)

        return list(
            sorted(
                set(
                    result_line.line
                    for result_line in self.lines.values()
                    if (result_line.get_score(query) <= best_score * 10)
                    or ResultLineType.CONTEXT in result_line.types
                )
            )
        )

    def get_result_blocks(self, query):
        lines_to_include = [
            line
            for line in sorted(self.lines.values(), key=lambda item: item.line)
            if line.line in self.get_lines(query)
        ]
        blocks = []

        for line in lines_to_include:
            distance_from_previous_line = (
                line.line - blocks[-1].lines[-1].line if blocks else 0
            )

            if not blocks or distance_from_previous_line > 1:
                blocks.append(ResultBlock(lines=[]))

            blocks[-1].lines.append(line)

        return blocks

    def to_json(self, query: str):
        return {
            "path": self.path,
            "fullPath": str(self.full_path),
            "score": round(self.get_best_score(query), 4),
            "blocks": [block.to_json(query) for block in self.get_result_blocks(query)],
        }

    def add_context_lines(self, lines: int):
        if lines == 0:
            return

        direction = lines // abs(lines)
        for result_line, _ in list(self.lines.items()):
            for offset in range(abs(lines)):
                new_line = result_line + (offset + 1) * direction

                if (new_line) not in range(len(self.line_texts)):
                    continue

                if new_line not in self.lines:
                    self.lines[new_line] = ResultLine(
                        line=new_line,
                        vector_distance=0.0,
                        line_text=self.line_texts[new_line - 1],
                        types={ResultLineType.CONTEXT},
                    )

                self.lines[new_line].add_type(ResultLineType.CONTEXT)



import re
import subprocess
from pathlib import Path

import orjson

from seagoat.repository import Repository
from seagoat.result import Result
from seagoat.sources.chroma import MAXIMUM_VECTOR_DISTANCE
from seagoat.utils.file_types import is_file_type_supported


def _fetch(query_text: str, path: str, limit: int):
    query_text = re.sub(r"\s+", "|", query_text)
    files = {}

    cmd = [
        "rg",
        "--json",
        "--max-count",
        str(limit),
        "--ignore-case",
        "--max-filesize",
        "200K",
        query_text,
        path,
    ]

    try:
        rg_output = subprocess.check_output(cmd, encoding="utf-8")
    except subprocess.CalledProcessError as exception:
        rg_output = exception.output

    for line in rg_output.splitlines():
        result = orjson.loads(line)

        if result.get("type") == "match":
            result_data = result["data"]
            absolute_path = result_data["path"]["text"]
            relative_path = Path(absolute_path).relative_to(path)
            line_number = int(result_data["line_number"])

            if not is_file_type_supported(relative_path):
                continue

            if relative_path not in files:
                files[relative_path] = Result(str(relative_path), absolute_path)

            # This is so that ripgrep results are on comparable levels with chroma results
            files[relative_path].add_line(line_number, MAXIMUM_VECTOR_DISTANCE * 0.8)

    return files.values()


def initialize(repository: Repository):
    path = repository.path

    def fetch(query_text: str, limit: int):
        return _fetch(query_text, str(path), limit)

    def cache_chunk(_):
        # Ripgrep does not need a cache for chunks
        pass

    return {
        "fetch": fetch,
        "cache_chunk": cache_chunk,
    }
